# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Shell related functions."""

import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile

from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

try:
  import psutil
except ImportError:
  psutil = None

_DEFAULT_LOW_DISK_SPACE_THRESHOLD = 5 * 1024 * 1024 * 1024  # 5 GB.
_TRUSTED_HOST_LOW_DISK_SPACE_THRESHOLD = 2 * 1024 * 1024  # 2 GB.
FILE_COPY_BUFFER_SIZE = 10 * 1024 * 1024  # 10 MB.
HANDLE_OUTPUT_FILE_TYPE_REGEX = re.compile(
    br'.*pid:\s*(\d+)\s*type:\s*File\s*([a-fA-F0-9]+):\s*(.*)')

_system_temp_dir = None


def _low_disk_space_threshold():
  """Get the low disk space threshold."""
  if environment.is_trusted_host(ensure_connected=False):
    # Trusted hosts can run with less free space as they do not store builds or
    # corpora.
    return _TRUSTED_HOST_LOW_DISK_SPACE_THRESHOLD

  return _DEFAULT_LOW_DISK_SPACE_THRESHOLD


def copy_file(source_file_path, destination_file_path):
  """Faster version of shutil.copy with buffer size."""
  if not os.path.exists(source_file_path):
    logs.log_error('Source file %s for copy not found.' % source_file_path)
    return False

  error_occurred = False
  try:
    with open(source_file_path, 'rb') as source_file_handle:
      with open(destination_file_path, 'wb') as destination_file_handle:
        shutil.copyfileobj(source_file_handle, destination_file_handle,
                           FILE_COPY_BUFFER_SIZE)
  except:
    error_occurred = True

  # Make sure that the destination file actually exists.
  error_occurred |= not os.path.exists(destination_file_path)

  if error_occurred:
    logs.log_warn('Failed to copy source file %s to destination file %s.' %
                  (source_file_path, destination_file_path))
    return False

  return True


def clear_build_directory():
  """Clears the build directory."""
  remove_directory(environment.get_value('BUILDS_DIR'), recreate=True)


def clear_build_urls_directory():
  """Clears the build url directory."""
  remove_directory(environment.get_value('BUILD_URLS_DIR'), recreate=True)

  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.clear_build_urls_directory()


def clear_crash_stacktraces_directory():
  """Clears the crash stacktraces directory."""
  remove_directory(
      environment.get_value('CRASH_STACKTRACES_DIR'), recreate=True)


def clear_common_data_bundles_directory():
  """Clear the common data bundle directory."""
  remove_directory(environment.get_value('FUZZ_DATA'), recreate=True)


def clear_data_bundles_directory():
  """Clears the data bundles directory."""
  remove_directory(environment.get_value('DATA_BUNDLES_DIR'), recreate=True)


def clear_mutator_plugins_directory():
  """Clears the mutator plugins directory."""
  remove_directory(environment.get_value('MUTATOR_PLUGINS_DIR'), recreate=True)


def clear_data_directories():
  """Clear all data directories."""
  clear_build_directory()
  clear_build_urls_directory()
  clear_crash_stacktraces_directory()
  clear_common_data_bundles_directory()
  clear_data_bundles_directory()
  clear_fuzzers_directories()
  clear_temp_directory()
  clear_testcase_directories()
  clear_mutator_plugins_directory()

  persistent_cache.clear_values(clear_all=True)


def clear_data_directories_on_low_disk_space():
  """Clear all data directories on low disk space. This should ideally never
  happen, but when it does, we do this to keep the bot working in sane state."""
  free_disk_space = get_free_disk_space()
  if free_disk_space is None:
    # Can't determine free disk space, bail out.
    return

  if free_disk_space >= _low_disk_space_threshold():
    return

  logs.log_warn(
      'Low disk space detected, clearing all data directories to free up space.'
  )
  clear_data_directories()


def clear_device_temp_directories():
  """Clear device specific temp directories."""
  if environment.is_android() and environment.get_value('ANDROID_SERIAL'):
    from clusterfuzz._internal.platforms import android
    android.device.clear_temp_directories()


def clear_fuzzers_directories():
  """Clears the fuzzers directory."""
  remove_directory(environment.get_value('FUZZERS_DIR'), recreate=True)


def clear_temp_directory(clear_user_profile_directories=True):
  """Clear the temporary directories."""
  temp_directory = environment.get_value('BOT_TMPDIR')
  remove_directory(temp_directory, recreate=True)

  test_temp_directory = environment.get_value('TEST_TMPDIR')
  if test_temp_directory != temp_directory:
    remove_directory(test_temp_directory, recreate=True)

  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.clear_temp_directory()

  if not clear_user_profile_directories:
    return

  user_profile_root_directory = environment.get_value('USER_PROFILE_ROOT_DIR')
  if not user_profile_root_directory:
    return

  remove_directory(user_profile_root_directory, recreate=True)


@environment.local_noop
def clear_system_temp_directory():
  """Clear system specific temp directory."""

  def _delete_object(path, delete_func):
    """Delete a object with its delete function, ignoring any error."""
    try:
      delete_func(path)
    except:
      pass

  if environment.get_value('SKIP_SYSTEM_TEMP_CLEANUP'):
    # This provides a way to avoid clearing system temporary directory when it
    # can interfere with other processes on the system.
    return

  # Cache system temp directory to avoid iterating through the system dir list
  # on every gettempdir call. Also, it helps to avoid a case where temp dir
  # fills up the disk and gets ignored by gettempdir.
  global _system_temp_dir
  if not _system_temp_dir:
    _system_temp_dir = tempfile.gettempdir()

  # Use a custom cleanup rather than using |remove_directory| since it
  # recreates the directory and can mess up permissions and symlinks.
  for root, dirs, files in walk(_system_temp_dir, topdown=False):
    for name in files:
      _delete_object(os.path.join(root, name), os.remove)

    for name in dirs:
      _delete_object(os.path.join(root, name), os.rmdir)
  logs.log('Cleared system temp directory: %s' % _system_temp_dir)


def clear_testcase_directories():
  """Clears the testcase directories."""
  remove_directory(environment.get_value('FUZZ_INPUTS'), recreate=True)
  remove_directory(environment.get_value('FUZZ_INPUTS_DISK'), recreate=True)

  if environment.is_android() and environment.get_value('ANDROID_SERIAL'):
    from clusterfuzz._internal.platforms import android
    android.device.clear_testcase_directory()
  if environment.platform() == 'FUCHSIA':
    from clusterfuzz._internal.platforms import fuchsia
    fuchsia.device.clear_testcase_directory()
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.clear_testcase_directories()


def close_open_file_handles_if_needed(path):
  """Try to close all open file handle for a specific path."""
  if environment.platform() != 'WINDOWS':
    # Handle closing is only applicable on Windows platform.
    return

  resources_directory = environment.get_platform_resources_directory()
  handle_executable_path = os.path.join(resources_directory, 'handle.exe')
  handle_output = execute_command(
      '%s -accepteula "%s"' % (handle_executable_path, path))
  for line in handle_output.splitlines():
    match = HANDLE_OUTPUT_FILE_TYPE_REGEX.match(line)
    if not match:
      continue

    process_id = match.group(1).decode('utf-8')
    file_handle_id = match.group(2).decode('utf-8')
    file_path = match.group(3).decode('utf-8')

    logs.log(
        'Closing file handle id %s for path %s.' % (file_handle_id, file_path))
    execute_command('%s -accepteula -c %s -p %s -y' %
                    (handle_executable_path, file_handle_id, process_id))


def create_directory(directory, create_intermediates=False, recreate=False):
  """Creates |directory|. Create intermediate directories if
  |create_intermediates|. Ignore if it already exists and |recreate| is
   False."""
  if os.path.exists(directory):
    if recreate:
      remove_directory(directory)
    else:
      return True

  try:
    if create_intermediates:
      os.makedirs(directory)
    else:
      os.mkdir(directory)
  except:
    logs.log_error('Unable to create directory %s.' % directory)
    return False

  return True


def execute_command(shell_command):
  """Run a command, returning its output."""
  try:
    process_handle = subprocess.Popen(
        shell_command,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    output, _ = process_handle.communicate()
  except:
    logs.log_error('Error while executing command %s.' % shell_command)
    return ''

  return output


def get_command(command_line):
  """Get the command to pass to subprocess."""
  if environment.platform() == 'WINDOWS':
    return command_line

  return shlex.split(command_line, posix=True)


def get_command_line_from_argument_list(argument_list):
  """Convert a list of arguments to a string."""
  return subprocess.list2cmdline(argument_list)


def get_directory_file_count(directory_path):
  """Returns number of files within a directory (recursively)."""
  file_count = 0
  for (root, _, files) in walk(directory_path):
    for filename in files:
      file_path = os.path.join(root, filename)
      if not os.path.isfile(file_path):
        continue
      file_count += 1

  return file_count


def get_directory_size(directory_path):
  """Returns size of a directory (in bytes)."""
  directory_size = 0
  for (root, _, files) in walk(directory_path):
    for filename in files:
      file_path = os.path.join(root, filename)
      directory_size += os.path.getsize(file_path)

  return directory_size


def get_files_list(directory_path):
  """Returns a list of files in a directory (recursively)."""
  files_list = []
  for (root, _, files) in walk(directory_path):
    for filename in files:
      file_path = os.path.join(root, filename)
      if not os.path.isfile(file_path):
        continue
      files_list.append(file_path)

  return files_list


def get_free_disk_space(path='/'):
  """Return free disk space."""
  if not os.path.exists(path):
    return None

  return psutil.disk_usage(path).free


def get_interpreter(file_to_execute):
  """Gives the interpreter needed to execute |file_to_execute|."""
  interpreters = {
      '.bash': 'bash',
      '.class': 'java',
      '.js': 'node',
      '.pl': 'perl',
      '.py': sys.executable,
      '.pyc': sys.executable,
      '.sh': 'sh'
  }

  try:
    interpreter = interpreters[os.path.splitext(file_to_execute)[1]]
  except KeyError:
    return None

  return interpreter


def get_execute_command(file_to_execute):
  """Return command to execute |file_to_execute|."""
  interpreter_path = get_interpreter(file_to_execute)

  # Hack for Java scripts.
  file_to_execute = file_to_execute.replace('.class', '')

  if interpreter_path:
    command = '%s %s' % (interpreter_path, file_to_execute)
  else:
    # Handle executables that don't need an interpreter.
    command = file_to_execute
  return command


def move(src, dst):
  """Wrapper around shutil.move(src, dst). If shutil.move throws an shutil.Error
  the exception is caught, an error is logged, and False is returned."""
  try:
    shutil.move(src, dst)
    return True
  except shutil.Error:
    logs.log_error('Failed to move %s to %s' % (src, dst))
    return False


def remove_empty_files(root_path):
  """Removes empty files in a path recursively"""
  for directory, _, filenames in walk(root_path):
    for filename in filenames:
      path = os.path.join(directory, filename)
      if os.path.getsize(path) > 0:
        continue

      try:
        os.remove(path)
      except:
        logs.log_error('Unable to remove the empty file: %s (%s).' %
                       (path, sys.exc_info()[0]))


def remove_empty_directories(path):
  """Removes empty folder in a path recursively."""
  if not os.path.isdir(path):
    return

  # Remove empty sub-folders.
  files = os.listdir(path)
  for filename in files:
    absolute_path = os.path.join(path, filename)
    if os.path.isdir(absolute_path):
      remove_empty_directories(absolute_path)

  # If folder is empty, delete it.
  files = os.listdir(path)
  if not files:
    try:
      os.rmdir(path)
    except:
      logs.log_error('Unable to remove empty folder %s.' % path)


def remove_file(file_path):
  """Removes a file, ignoring any error if it occurs."""
  try:
    if os.path.exists(file_path):
      os.remove(file_path)
  except:
    pass


def remove_directory(directory, recreate=False, ignore_errors=False):
  """Removes a directory tree."""
  # Log errors as warnings if |ignore_errors| is set.
  log_error_func = logs.log_warn if ignore_errors else logs.log_error

  def clear_read_only(func, path, _):
    """Clear the read-only bit and reattempt the removal again.
    This is needed on Windows."""

    try:
      os.chmod(path, 0o750)
    except:
      # If this is tmpfs, we will probably fail.
      pass

    try:
      func(path)
    except:
      # Log errors for all cases except device or resource busy errors, as such
      # errors are expected in cases when mounts are used.
      error_message = str(sys.exc_info()[1])
      if 'Device or resource busy' not in error_message:
        logs.log_warn(
            'Failed to remove directory %s failed because %s with %s failed. %s'
            % (directory, func, path, error_message))

  # Try the os-specific deletion commands first. This helps to overcome issues
  # with unicode filename handling.
  if os.path.exists(directory):
    if environment.platform() == 'WINDOWS':
      os.system('rd /s /q "%s" > nul 2>&1' % directory)
    else:
      os.system('rm -rf "%s" > /dev/null 2>&1' % directory)

  if os.path.exists(directory):
    # If the directory still exists after using native OS delete commands, then
    # try closing open file handles and then try removing it with read only
    # bit removed (Windows only).
    close_open_file_handles_if_needed(directory)
    shutil.rmtree(directory, onerror=clear_read_only)

  if os.path.exists(directory):
    # 1. If directory is a mount point, then directory itself won't be
    #    removed. So, check the list of files inside it.
    # 2. If directory is a regular directory, then it should have not
    #    existed.
    if not os.path.ismount(directory) or os.listdir(directory):
      # Directory could not be cleared. Bail out.
      log_error_func('Failed to clear directory %s.' % directory)
      return False

    return True

  if not recreate:
    return True

  try:
    os.makedirs(directory)
  except:
    log_error_func('Unable to re-create directory %s.' % directory)
    return False

  return True


def walk(directory, **kwargs):
  """Wrapper around walk to resolve compatibility issues."""
  return os.walk(directory, **kwargs)


# Copy of shutil.which from Python 3.3 (unavailable in Python 2.7).
# pylint: disable=bad-inline-option,g-inconsistent-quotes,redefined-builtin
# yapf: disable
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
  """Given a command, mode, and a PATH string, return the path which
  conforms to the given mode on the PATH, or None if there is no such
  file.
  `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
  of os.environ.get("PATH"), or can be overridden with a custom search
  path.
  Note: This function was backported from the Python 3 source code.
  """

  # Check that a given file can be accessed with the correct mode.
  # Additionally check that `file` is not a directory, as on Windows
  # directories pass the os.access check.

  def _access_check(fn, mode):
    return (
        os.path.exists(fn) and os.access(fn, mode) and not os.path.isdir(fn)
    )

  # If we're given a path with a directory part, look it up directly
  # rather than referring to PATH directories. This includes checking
  # relative to the current directory, e.g. ./script
  if os.path.dirname(cmd):
    if _access_check(cmd, mode):
      return cmd

    return None

  if path is None:
    path = os.environ.get("PATH", os.defpath)
  if not path:
    return None

  path = path.split(os.pathsep)

  if sys.platform == "win32":
    # The current directory takes precedence on Windows.
    if os.curdir not in path:
      path.insert(0, os.curdir)

    # PATHEXT is necessary to check on Windows.
    pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
    # See if the given file matches any of the expected path
    # extensions. This will allow us to short circuit when given
    # "python.exe". If it does match, only test that one, otherwise we
    # have to try others.
    if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
      files = [cmd]
    else:
      files = [cmd + ext for ext in pathext]
  else:
    # On other platforms you don't have things like PATHEXT to tell you
    # what file suffixes are executable, so just pass on cmd as-is.
    files = [cmd]

  seen = set()
  for dir in path:
    normdir = os.path.normcase(dir)
    if normdir not in seen:
      seen.add(normdir)
      for thefile in files:
        name = os.path.join(dir, thefile)
        if _access_check(name, mode):
          return name

  return None
# pylint: enable=bad-inline-option,g-inconsistent-quotes,redefined-builtin
# yapf: enable



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Battery functions."""

import datetime
import re
import time

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

from . import adb
from . import settings

BATTERY_CHARGE_INTERVAL = 30 * 60  # 0.5 hour.
BATTERY_CHECK_INTERVAL = 15 * 60  # 15 minutes.
EXPECTED_BATTERY_LEVEL = 60  # A percentage.
EXPECTED_BATTERY_TEMPERATURE = 35.0  # Degrees Celsius.
LOW_BATTERY_LEVEL_THRESHOLD = 30  # A percentage.
MAX_BATTERY_TEMPERATURE_THRESHOLD = 37.0  # Don't change this or battery swells.

LAST_BATTERY_CHECK_TIME_KEY = 'android_last_battery_check'


def get_battery_level_and_temperature():
  """Return device's battery and temperature levels."""
  output = adb.run_shell_command(['dumpsys', 'battery'])

  # Get battery level.
  m_battery_level = re.match(r'.*level: (\d+).*', output, re.DOTALL)
  if not m_battery_level:
    logs.log_error('Error occurred while getting battery status.')
    return None

  # Get battery temperature.
  m_battery_temperature = re.match(r'.*temperature: (\d+).*', output, re.DOTALL)
  if not m_battery_temperature:
    logs.log_error('Error occurred while getting battery temperature.')
    return None

  level = int(m_battery_level.group(1))
  temperature = float(m_battery_temperature.group(1)) / 10.0
  return {'level': level, 'temperature': temperature}


def wait_until_good_state():
  """Check battery and make sure it is charged beyond minimum level and
  temperature thresholds."""
  # Battery levels are not applicable on GCE.
  if environment.is_android_cuttlefish() or settings.is_automotive():
    return

  # Make sure device is online.
  adb.wait_for_device()

  # Skip battery check if done recently.
  last_battery_check_time = persistent_cache.get_value(
      LAST_BATTERY_CHECK_TIME_KEY,
      constructor=datetime.datetime.utcfromtimestamp)
  if last_battery_check_time and not dates.time_has_expired(
      last_battery_check_time, seconds=BATTERY_CHECK_INTERVAL):
    return

  # Initialize variables.
  battery_level_threshold = environment.get_value('LOW_BATTERY_LEVEL_THRESHOLD',
                                                  LOW_BATTERY_LEVEL_THRESHOLD)
  battery_temperature_threshold = environment.get_value(
      'MAX_BATTERY_TEMPERATURE_THRESHOLD', MAX_BATTERY_TEMPERATURE_THRESHOLD)
  device_restarted = False

  while True:
    battery_information = get_battery_level_and_temperature()
    if battery_information is None:
      logs.log_error('Failed to get battery information, skipping check.')
      return

    battery_level = battery_information['level']
    battery_temperature = battery_information['temperature']
    logs.log('Battery information: level (%d%%), temperature (%.1f celsius).' %
             (battery_level, battery_temperature))
    if (battery_level >= battery_level_threshold and
        battery_temperature <= battery_temperature_threshold):
      persistent_cache.set_value(LAST_BATTERY_CHECK_TIME_KEY, time.time())
      return

    logs.log('Battery in bad battery state, putting device in sleep mode.')

    if not device_restarted:
      adb.reboot()
      device_restarted = True

    # Change thresholds to expected levels (only if they were below minimum
    # thresholds).
    if battery_level < battery_level_threshold:
      battery_level_threshold = EXPECTED_BATTERY_LEVEL
    if battery_temperature > battery_temperature_threshold:
      battery_temperature_threshold = EXPECTED_BATTERY_TEMPERATURE

    # Stopping shell should help with shutting off a lot of services that would
    # otherwise use up the battery. However, we need to turn it back on to get
    # battery status information.
    adb.stop_shell()
    time.sleep(BATTERY_CHARGE_INTERVAL)
    adb.start_shell()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Issue tracker manager functions."""

import json

import jira

from clusterfuzz._internal.config import db_config


class IssueTrackerManager(object):
  """Issue tracker manager."""

  def __init__(self, project_name):
    """"Construct an issue tracker manager instance based on parameters."""
    self._client = None
    self.project_name = project_name

  @property
  def client(self):
    """HTTP Client."""
    if self._client is None:
      self._client = self._create_client()

    return self._client

  def _create_client(self):
    """Return a client object for querying the issue tracker."""
    config = db_config.get()
    credentials = json.loads(config.jira_credentials)
    jira_url = config.jira_url
    jira_client = jira.JIRA(
        jira_url, auth=(credentials['username'], credentials['password']))
    return jira_client

  def save(self, issue):
    """Save an issue."""
    self._update(issue)

  def create(self):
    """Create an issue."""
    default_fields = {
        'project': self.project_name,
        'summary': 'Default summary',
        'description': 'Default description',
        'issuetype': {
            'name': 'Bug'
        }
    }

    jira_issue = self.client.create_issue(fields=default_fields)
    return jira_issue

  def _update(self, issue):
    """Update an issue."""

    # Get labels from LabelStore.
    labels = list(issue.labels)

    # Get components from LabelStore.
    components = list(issue.components)

    # Get watchers from LabelStore.
    watchers = list(issue.ccs)

    update_fields = {
        'summary': issue.title,
        'description': issue.body,
        'labels': labels,
        'components': components,
    }

    # Only add status if it has changed.
    # Brittle - we should be pulling the equivalent of 'new' from the policy.
    if issue.status != 'Open':
      status = {'name': issue.status}
      update_fields['status'] = status

    if issue.assignee is not None:
      assignee = {'name': issue.assignee}
      update_fields['assignee'] = assignee

    # Again brittle - need to pull these strings from policy.
    if 'Critical - P1' in labels:
      update_fields['priority'] = {'name': 'Critical - P1'}
    elif 'Major - P2' in labels:
      update_fields['priority'] = {'name': 'Major - P2'}

    # Jira weirdness, update watchers this way.
    for watcher in watchers:
      self.client.add_watcher(issue.jira_issue, watcher)

    issue.jira_issue.update(fields=update_fields)

  def get_watchers(self, issue):
    """Retrieve list of watchers."""
    watchlist = self.client.watchers(issue)
    watchers = []
    for watcher in watchlist.watchers:
      watchers.append(watcher.name)
    return watchers

  def get_issue(self, issue_id):
    """Retrieve an issue object with a specific id."""
    issue = self.client.issue(str(issue_id))
    return issue

  def get_issue_count(self, query_string):
    """Return number of issues for a given query."""
    issues = self.client.search_issues(query_string)
    return len(issues)

  def get_issues(self, query_string, max_results=10000):
    """Return all issues for a given query."""
    issues = self.client.search_issues(query_string, maxResults=max_results)
    return issues



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Getting and using custom mutator plugins for libFuzzer."""

import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

MUTATOR_SHARED_OBJECT_FILENAME = 'mutator-plugin.so'
ARCHIVES_SUBDIR_NAME = 'archives'
PLUGINS_SUBDIR_NAME = 'plugins'


def _get_mutator_plugins_bucket_url():
  """Returns the url of the mutator plugin's cloud storage bucket."""
  mutator_plugins_bucket = environment.get_value('MUTATOR_PLUGINS_BUCKET')
  if not mutator_plugins_bucket:
    logs.log_warn('MUTATOR_PLUGINS_BUCKET is not set in project config, '
                  'skipping custom mutator strategy.')
    return None

  return 'gs://%s' % mutator_plugins_bucket


def _get_mutator_plugins_subdir(subdir):
  """Returns the path of the subdirectory named |subdir| in
  MUTATOR_PLUGINS_DIR."""
  return os.path.join(environment.get_value('MUTATOR_PLUGINS_DIR'), subdir)


def _get_mutator_plugins_archives_dir():
  """Returns path of the archives subdirectory in MUTATOR_PLUGINS_DIR"""
  return _get_mutator_plugins_subdir(ARCHIVES_SUBDIR_NAME)


def _get_mutator_plugins_unpacked_dir():
  """Returns path of the unpacked plugins subdirectory in MUTATOR_PLUGINS_DIR"""
  return _get_mutator_plugins_subdir(PLUGINS_SUBDIR_NAME)


def _get_mutator_plugins_from_bucket():
  """Returns list of the mutator plugin archives in the mutator plugin storage
  bucket."""
  mutator_plugins_bucket_url = _get_mutator_plugins_bucket_url()
  if not mutator_plugins_bucket_url:
    return None

  return storage.list_blobs(mutator_plugins_bucket_url)


def _download_mutator_plugin_archive(mutator_plugin_archive):
  """Downloads the |mutator_plugin_archive| from the mutator plugin storage
  bucket to the plugin archives directory. Returns the path that the archive was
  downloaded to."""
  file_path = os.path.join(_get_mutator_plugins_archives_dir(),
                           mutator_plugin_archive)
  url = '%s/%s' % (_get_mutator_plugins_bucket_url(), mutator_plugin_archive)
  if not storage.copy_file_from(url, file_path):
    logs.log_error(
        'Failed to copy plugin archive from %s to %s' % (url, file_path))
    return None

  return file_path


def _unpack_mutator_plugin(mutator_plugin_archive_path):
  """Unpacks |mutator_plugin_archive_path| in the unpacked plugins directory and
  returns the path it was unpacked into."""
  mutator_plugin_name = os.path.basename(
      os.path.splitext(mutator_plugin_archive_path)[0])
  unpacked_plugin_dir = os.path.join(_get_mutator_plugins_unpacked_dir(),
                                     mutator_plugin_name)
  archive.unpack(mutator_plugin_archive_path, unpacked_plugin_dir)
  return unpacked_plugin_dir


def _extract_name_from_archive(plugin_archive_filename):
  """Parses |plugin_archive_filename| which should be named using the schema:
  '$NAME-$JOB-$FUZZ_TARGET.zip'. Returns tuple containing
  ($NAME, $JOB-$FUZZ_TARGET)."""
  # TODO(metzman): Get rid of this when we create an upload page for custom
  # mutator plugins.
  plugin_archive_name = os.path.splitext(plugin_archive_filename)[0]
  idx = plugin_archive_name.index('-')
  plugin_name = plugin_archive_name[:idx]
  return plugin_name, plugin_archive_name[idx + 1:]


class PluginGetter(object):
  """Class that gets a usable mutator plugin for |fuzzer_binary_name| from
  GCS."""

  def __init__(self, fuzzer_binary_name):
    self.fuzzer_binary_name = fuzzer_binary_name
    self.job_name = environment.get_value('JOB_NAME')
    self.create_directories()

  def create_directories(self):
    """Creates directories needed to use mutator plugins."""
    # TODO(320): Change mutator plugin downloads so that they don't need to be
    # deleted and redownloaded on each run of launcher.py.
    shell.create_directory(
        environment.get_value('MUTATOR_PLUGINS_DIR'),
        create_intermediates=True,
        recreate=True)
    shell.create_directory(
        _get_mutator_plugins_archives_dir(),
        create_intermediates=True,
        recreate=True)
    shell.create_directory(
        _get_mutator_plugins_unpacked_dir(),
        create_intermediates=True,
        recreate=True)

  def _is_plugin_usable(self, plugin_archive_filename):
    """Returns True if |plugin_archive_filename| is a usable plugin for this
    job, fuzz target combination."""
    _, plugin_job_and_fuzzer = _extract_name_from_archive(
        plugin_archive_filename)
    expected_name = '%s-%s' % (self.job_name, self.fuzzer_binary_name)
    return expected_name == plugin_job_and_fuzzer

  def get_mutator_plugin(self):
    """Downloads and unpacks a usable mutator plugin for this job and fuzz
    target if one is available in GCS"""
    mutator_plugins = _get_mutator_plugins_from_bucket()
    if not mutator_plugins:
      # No plugins found or plugin url is not set.
      return None

    usable_mutator_plugins = [
        plugin_archive for plugin_archive in mutator_plugins
        if self._is_plugin_usable(plugin_archive)
    ]

    # Quit if no usable plugins are available.
    if not usable_mutator_plugins:
      return None

    plugin_archive_name = utils.random_element_from_list(usable_mutator_plugins)

    # Handle a failed download.
    plugin_archive_path = _download_mutator_plugin_archive(plugin_archive_name)
    if not plugin_archive_path:
      return None

    _unpack_mutator_plugin(plugin_archive_path)
    mutator_plugin_path = find_mutator_plugin()
    if mutator_plugin_path is None:
      logs.log_error('Could not find plugin in %s' % plugin_archive_path)

    return mutator_plugin_path


def find_mutator_plugin():
  """Sets LD_PRELOAD to the path of a usable mutator plugin shared object.
  This should only be called after a call to get_mutator_plugin."""
  paths = shell.get_files_list(_get_mutator_plugins_unpacked_dir())
  # This function should not be called unless there is an unpacked plugin.
  for path in paths:
    if os.path.basename(path) == MUTATOR_SHARED_OBJECT_FILENAME:
      return path
  return None


def get_mutator_plugin(fuzzer_binary_name):
  """Downloads and unpacks a mutator plugin if a usable one for
  |fuzzer_binary_name| is in the bucket."""
  plugin_getter = PluginGetter(fuzzer_binary_name)
  return plugin_getter.get_mutator_plugin()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer stats schema."""

_COMMON_COLUMNS = [{
    'mode': 'NULLABLE',
    'name': 'kind',
    'type': 'STRING'
}, {
    'mode': 'NULLABLE',
    'name': 'timestamp',
    'type': 'FLOAT'
}, {
    'mode': 'REPEATED',
    'name': 'command',
    'type': 'STRING'
}, {
    'mode': 'NULLABLE',
    'name': 'build_revision',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'fuzzer',
    'type': 'STRING'
}, {
    'mode': 'NULLABLE',
    'name': 'job',
    'type': 'string'
}, {
    'mode': 'NULLABLE',
    'name': 'source',
    'type': 'STRING'
}]

_AFL_SCHEMA = [{
    'mode': 'NULLABLE',
    'name': 'dict_used',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'new_units_generated',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'crash_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'manual_dict_size',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'timeout_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'stability',
    'type': 'FLOAT'
}, {
    'mode': 'NULLABLE',
    'name': 'corpus_crash_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'startup_crash_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'timeout_limit',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_subset',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'bad_instrumentation',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'average_exec_per_sec',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'corpus_size',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'new_units_added',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'log_lines_unwanted',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'actual_duration',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'expected_duration',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'oom_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'log_lines_ignored',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_mutations',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'read_units',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'max_len',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'log_lines_from_engine',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_recommended_dict',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'exception_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'slow_unit_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_value_profile',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'recommended_dict_size',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'leak_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_mutations_radamsa',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_mutations_ml_rnn',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_selection_method',
    'type': 'STRING'
}] + _COMMON_COLUMNS

_LIBFUZZER_SCHEMA = [{
    'mode': 'NULLABLE',
    'name': 'max_len',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'log_lines_ignored',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'feature_coverage',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'average_exec_per_sec',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'crash_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'expected_duration',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'leak_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_subset',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_dataflow_tracing',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'number_of_executed_units',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'slow_unit_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_mutations_radamsa',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'log_lines_from_engine',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'merge_edge_coverage',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_random_max_len',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'peak_rss_mb',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'corpus_crash_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'startup_crash_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'edge_coverage',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'new_units_generated',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'initial_feature_coverage',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'manual_dict_size',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_fork',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'bad_instrumentation',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'initial_edge_coverage',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'timeout_limit',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'dict_used',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'fuzzing_time_percent',
    'type': 'FLOAT'
}, {
    'name': 'new_features',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'log_lines_unwanted',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'slowest_unit_time_sec',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'edges_total',
    'type': 'INTEGER'
}, {
    'name': 'new_edges',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'timeout_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_mutations_ml_rnn',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_entropic',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'slow_units_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'oom_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_selection_method',
    'type': 'STRING'
}, {
    'mode': 'NULLABLE',
    'name': 'recommended_dict_size',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'actual_duration',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_value_profile',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_recommended_dict',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'corpus_size',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'new_units_added',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_mutator_plugin',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_mutator_plugin_radamsa',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_peach_grammar_mutation',
    'type': 'STRING'
}, {
    'mode': 'NULLABLE',
    'name': 'corpus_rss_mb',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'merge_new_files',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'merge_new_features',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_handle_unstable',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_weighted_mutations',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'strategy_corpus_mutations',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'read_units',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'seed_corpus_rss',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'exception_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'edge_coverage_count',
    'type': 'INTEGER'
}] + _COMMON_COLUMNS

_HONGGFUZZ_SCHEMA = [{
    'mode': 'NULLABLE',
    'name': 'timeout_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'branch_coverage_percent',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'guard_nb',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'crashes_count',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'new_units_added',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'slowest_unit_ms',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'peak_rss_mb',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'iterations',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'time',
    'type': 'INTEGER'
}, {
    'mode': 'NULLABLE',
    'name': 'speed',
    'type': 'INTEGER'
}] + _COMMON_COLUMNS

_SCHEMA = {
    'afl': _AFL_SCHEMA,
    'honggfuzz': _HONGGFUZZ_SCHEMA,
    'libFuzzer': _LIBFUZZER_SCHEMA,
}


def get(engine_name):
  """Get the schema for an engine name."""
  schema = _SCHEMA.get(engine_name)
  if not schema:
    return None

  return {'fields': schema}



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for service_accounts.py"""

import unittest

from handlers.cron import service_accounts


class ServiceAccountIdTest(unittest.TestCase):
  """Tests _service_account_id."""

  def test_min_length(self):
    self.assertEqual(service_accounts._service_account_id('a'), 'bot-a0')  # pylint: disable=protected-access

  def test_regular_length(self):
    self.assertEqual(service_accounts._service_account_id('abc'), 'bot-abc')  # pylint: disable=protected-access

  def test_max_length(self):
    self.assertEqual(
        service_accounts._service_account_id('a' * 26), 'bot-%s' % ('a' * 26))  # pylint: disable=protected-access

  def test_more_than_max_length(self):
    self.assertEqual('bot-52a00b8461593ce33409d7c5d0',
                     service_accounts._service_account_id('a' * 27))  # pylint: disable=protected-access



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compute_engine_projects."""

import unittest

from clusterfuzz._internal.google_cloud_utils import compute_engine_projects


class LoadProjectTest(unittest.TestCase):
  """Tests load_project."""

  def test_load_test_project(self):
    """Test that test config (project test-clusterfuzz) loads without any
    exceptions."""
    self.assertIsNotNone(
        compute_engine_projects.load_project('test-clusterfuzz'))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for corpus tagging helper functions."""

import unittest

from clusterfuzz._internal.datastore import corpus_tagging
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class CorpusTaggingTest(unittest.TestCase):
  """Test corpus tagging helper functions."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_get_targets_with_correct_tag(self):
    """Tests get_targets_with_tag returns the correct fuzz target."""
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()

    self.assertEqual(
        'test_fuzz_target',
        corpus_tagging.get_targets_with_tag('test_tag')[0]
        .fully_qualified_fuzz_target_name)

  def test_get_no_results_with_incorrect_tag(self):
    """Tests get_targets_with_tag returns the nothing if non match the tag."""

    self.assertEqual(0, len(corpus_tagging.get_targets_with_tag('test_tag')))

  def test_get_all_targets_with_correct_tag(self):
    """Tests that get_targets_with tag returns all and only targets with
    matching tag."""
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target1').put()

    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target2').put()

    data_types.CorpusTag(
        tag='not_the_same_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target3').put()

    self.assertEqual(2, len(corpus_tagging.get_targets_with_tag('test_tag')))

  def test_get_target_with_correct_tag_when_target_has_multiple_tags(self):
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()
    data_types.CorpusTag(
        tag='test_tag2',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()

    self.assertEqual(
        'test_fuzz_target',
        corpus_tagging.get_targets_with_tag('test_tag')[0]
        .fully_qualified_fuzz_target_name)

  def test_get_tag_from_target_with_one_tag(self):
    """Test getting the tag of a given fuzz target."""
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()

    self.assertEqual(
        'test_tag',
        corpus_tagging.get_fuzz_target_tags('test_fuzz_target')[0].tag)

  def test_get_tags_from_target_with_multiple_tags(self):
    """Test getting the tags of a given fuzz target with more than one tag."""
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()
    data_types.CorpusTag(
        tag='test_tag2',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()

    tags = [
        i.tag for i in corpus_tagging.get_fuzz_target_tags('test_fuzz_target')
    ]

    self.assertEqual(tags, ['test_tag', 'test_tag2'])

  def test_get_similarly_tagged_fuzzers_returns_empty_on_no_matches(self):
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()
    data_types.CorpusTag(
        tag='test_tag2',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()

    self.assertEqual(
        {}, corpus_tagging.get_similarly_tagged_fuzzers('test_fuzz_target'))

  def test_get_similarly_tagged_fuzzers_returns_with_correct_values(self):
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target').put()
    data_types.CorpusTag(
        tag='test_tag',
        fully_qualified_fuzz_target_name='test_fuzz_target_2').put()

    self.assertEqual({
        'test_tag': ['test_fuzz_target_2']
    }, corpus_tagging.get_similarly_tagged_fuzzers('test_fuzz_target'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Minimizer based on the delta debugging algorithm."""

from . import errors
from . import minimizer
from . import utils


class DeltaTestcase(minimizer.Testcase):
  """Test case for the delta minimizer."""

  def _process_test_result(self, test_passed, hypothesis):
    """Update state based on test_passed and hypothesis."""
    # If we crashed or cannot split the test into smaller chunks, we're done.
    if not test_passed or len(hypothesis) <= 1:
      return

    # Test passed. Break this up into sub-tests.
    middle = len(hypothesis) // 2
    front = hypothesis[:middle]
    back = hypothesis[middle:]

    # Working back to front works better for some formats.
    self.prepare_test(back)
    self.prepare_test(front)


class DeltaMinimizer(minimizer.Minimizer):
  """Minimizer based on the delta algorithm."""

  def _execute(self, data):
    """Prepare tests for delta minimization and process."""
    testcase = DeltaTestcase(data, self)
    if not self.validate_tokenizer(data, testcase):
      raise errors.TokenizationFailureError('Delta Minimizer')

    tokens = testcase.tokens

    step = max(1, len(tokens) // self.max_threads)
    for start in range(0, len(tokens), step):
      end = min(len(tokens), start + step)
      hypothesis = list(range(start, end))
      testcase.prepare_test(hypothesis)

    testcase.process()
    return testcase

  @staticmethod
  def run(data, thread_count=minimizer.DEFAULT_THREAD_COUNT, file_extension=''):
    """Try to minimize |data| using a simple line tokenizer."""
    delta_minimizer = DeltaMinimizer(
        utils.test, max_threads=thread_count, file_extension=file_extension)
    return delta_minimizer.minimize(data)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for analyze task."""

import unittest

from clusterfuzz._internal.bot.tasks import analyze_task
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class AddDefaultIssueMetadataTest(unittest.TestCase):
  """Test _add_default_issue_metadata."""

  def setUp(self):
    helpers.patch(
        self,
        [
            'clusterfuzz._internal.bot.fuzzers.engine_common.'
            'get_all_issue_metadata_for_testcase',
            # Disable logging.
            'clusterfuzz._internal.datastore.data_types.Testcase._post_put_hook',
            'clusterfuzz._internal.metrics.logs.log',
        ])

  def test_union(self):
    """Test union of current testcase metadata and default issue metadata."""
    self.mock.get_all_issue_metadata_for_testcase.return_value = {
        'issue_owners': 'dev1@example1.com, dev2@example2.com',
        'issue_components': 'component1',
        'issue_labels': 'label1, label2 ,label3'
    }

    testcase = test_utils.create_generic_testcase()
    testcase.set_metadata('issue_owners', 'dev3@example3.com,dev2@example2.com')
    testcase.set_metadata('issue_components', 'component2')
    testcase.set_metadata('issue_labels', 'label4,label5, label2,')

    analyze_task._add_default_issue_metadata(testcase)  # pylint: disable=protected-access
    self.assertEqual('dev1@example1.com,dev2@example2.com,dev3@example3.com',
                     testcase.get_metadata('issue_owners'))
    self.assertEqual('component1,component2',
                     testcase.get_metadata('issue_components'))
    self.assertEqual('label1,label2,label3,label4,label5',
                     testcase.get_metadata('issue_labels'))
    self.assertEqual(3, self.mock.log.call_count)

  def test_no_testcase_metadata(self):
    """Test when we only have default issue metadata and no testcase
    metadata."""
    self.mock.get_all_issue_metadata_for_testcase.return_value = None

    testcase = test_utils.create_generic_testcase()
    testcase.set_metadata('issue_owners', 'dev1@example1.com,dev2@example2.com')
    testcase.set_metadata('issue_components', 'component1')
    testcase.set_metadata('issue_labels', 'label1,label2,label3')

    analyze_task._add_default_issue_metadata(testcase)  # pylint: disable=protected-access
    self.assertEqual('dev1@example1.com,dev2@example2.com',
                     testcase.get_metadata('issue_owners'))
    self.assertEqual('component1', testcase.get_metadata('issue_components'))
    self.assertEqual('label1,label2,label3',
                     testcase.get_metadata('issue_labels'))
    self.assertEqual(0, self.mock.log.call_count)

  def test_no_default_issue_metadata(self):
    """Test when we only have testcase metadata and no default issue
    metadata."""
    self.mock.get_all_issue_metadata_for_testcase.return_value = {
        'issue_owners': 'dev1@example1.com,dev2@example2.com',
        'issue_components': 'component1',
        'issue_labels': 'label1,label2,label3'
    }

    testcase = test_utils.create_generic_testcase()

    analyze_task._add_default_issue_metadata(testcase)  # pylint: disable=protected-access
    self.assertEqual('dev1@example1.com,dev2@example2.com',
                     testcase.get_metadata('issue_owners'))
    self.assertEqual('component1', testcase.get_metadata('issue_components'))
    self.assertEqual('label1,label2,label3',
                     testcase.get_metadata('issue_labels'))
    self.assertEqual(3, self.mock.log.call_count)

  def test_same_testcase_and_default_issue_metadata(self):
    """Test when we have same testcase metadata and default issue metadata."""
    self.mock.get_all_issue_metadata_for_testcase.return_value = {
        'issue_owners': 'dev1@example1.com,dev2@example2.com',
        'issue_components': 'component1',
        'issue_labels': 'label1,label2,label3'
    }

    testcase = test_utils.create_generic_testcase()
    testcase.set_metadata('issue_owners', 'dev1@example1.com,dev2@example2.com')
    testcase.set_metadata('issue_components', 'component1')
    testcase.set_metadata('issue_labels', 'label1,label2,label3')

    analyze_task._add_default_issue_metadata(testcase)  # pylint: disable=protected-access
    self.assertEqual('dev1@example1.com,dev2@example2.com',
                     testcase.get_metadata('issue_owners'))
    self.assertEqual('component1', testcase.get_metadata('issue_components'))
    self.assertEqual('label1,label2,label3',
                     testcase.get_metadata('issue_labels'))
    self.assertEqual(0, self.mock.log.call_count)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer related helper functions for testcases."""

import os
import random

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.metrics import logs

# FIXME: Importing data_handler module is heavyweight and
# runs into exception with |oauth2client|. Skip and initialize
# constant directly.
LOCK_FILENAME = '.lock'


def create_testcase_list_file(testcase_file_paths, input_directory):
  """Store list of fuzzed testcases from fuzzer in a bot specific
  testcase list file."""
  if not testcase_file_paths:
    logs.log_error('No testcases found, skipping list file.')
    return

  bot_testcases_file_path = utils.get_bot_testcases_file_path(input_directory)
  with open(bot_testcases_file_path, 'wb') as bot_testcases_file_handle:
    bot_testcases_file_handle.write('\n'.join(testcase_file_paths))


def is_valid_testcase_file(file_path,
                           check_if_exists=True,
                           size_limit=None,
                           allowed_extensions=None):
  """Return true if the file looks like a testcase file."""
  return utils.is_valid_testcase_file(file_path, check_if_exists, size_limit,
                                      allowed_extensions)


def is_locked(input_directory):
  """Returns true if the data bundle is locked and unavailable for use."""
  lock_file_path = os.path.join(input_directory, LOCK_FILENAME)
  return os.path.exists(lock_file_path)


def get_random_testcases(input_directory, max_testcases):
  """Returns list of |max_testcases| testcases."""
  testcases_list = get_testcases(input_directory)
  return random.SystemRandom().sample(testcases_list, max_testcases)


def get_testcases(input_directory):
  """Returns list of testcase files."""
  testcase_list_file_path = os.path.join(
      input_directory, testcase_manager.TESTCASE_LIST_FILENAME)
  if not os.path.exists(testcase_list_file_path):
    return []

  with open(testcase_list_file_path, 'rb') as testcase_list_file_handle:
    testcase_relative_file_paths = testcase_list_file_handle.read().splitlines()

  testcase_file_paths = []
  for testcase_relative_file_path in testcase_relative_file_paths:
    # Discard junk paths.
    if not testcase_relative_file_path.strip():
      continue

    testcase_file_path = os.path.join(
        input_directory, testcase_relative_file_path.replace('/', os.sep))
    testcase_file_paths.append(testcase_file_path)

  return testcase_file_paths



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JSON helper utilities."""

import datetime
import json


class JSONEncoder(json.JSONEncoder):
  """Custom version of JSON encoder with support for additional object types
  (e.g. datetime)."""

  def default(self, o):  # pylint: disable=method-hidden
    if isinstance(o, datetime.datetime):
      return {
          '__type__': 'datetime',
          'year': o.year,
          'month': o.month,
          'day': o.day,
          'hour': o.hour,
          'minute': o.minute,
          'second': o.second,
          'microsecond': o.microsecond,
      }
    if isinstance(o, datetime.date):
      return {
          '__type__': 'date',
          'year': o.year,
          'month': o.month,
          'day': o.day,
      }

    return json.JSONEncoder.default(self, o)


class JSONDecoder(json.JSONDecoder):
  """Custom version of JSON decoder with support for additional object types
  (e.g. datetime)."""
  _TYPES = {'datetime': datetime.datetime, 'date': datetime.date}

  def __init__(self, *args, **kwargs):
    super(JSONDecoder, self).__init__(
        object_hook=self.dict_to_object, *args, **kwargs)

  def dict_to_object(self, d):
    if '__type__' not in d:
      return d

    object_type = d.pop('__type__')
    try:
      return self._TYPES[object_type](**d)
    except Exception:
      d['__type__'] = object_type
      return d


def dumps(obj, *args, **kwargs):
  """Custom json.dumps using custom encoder JSONEncoder defined in this file."""
  kwargs['cls'] = JSONEncoder
  kwargs['sort_keys'] = True
  return json.dumps(obj, *args, **kwargs)


def loads(obj, *args, **kwargs):
  """Custom json.loads using custom encoder JSONDecoder defined in this file."""
  kwargs['cls'] = JSONDecoder
  return json.loads(obj, *args, **kwargs)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for managing Google Cloud Storage."""

import copy
import datetime
import json
import os
import shutil
import threading
import time

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import locks
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

from . import credentials

try:
  import google.cloud
  from google.cloud import storage as gcs
except ImportError:
  # This is expected to fail on AppEngine.
  pass

# Usually, authentication time have expiry of ~30 minutes, but keeping this
# values lower to avoid failures and any future changes.
AUTH_TOKEN_EXPIRY_TIME = 10 * 60

# Wait time to let NFS copy to sync across bricks.
CACHE_COPY_WAIT_TIME = 10

# Cache directory name.
CACHE_DIRNAME = 'cache'

# Time to hold the cache lock for.
CACHE_LOCK_TIMEOUT = 30 * 60

# File extension for tmp cache files.
CACHE_METADATA_FILE_EXTENSION = '.metadata'

# Maximum size of file to allow in cache.
CACHE_SIZE_LIMIT = 5 * 1024 * 1024 * 1024  # 5 GB

# Cache get/set timeout.
CACHE_TIMEOUT = 3 * 60 * 60  # 3 hours.

# The number of retries to perform some GCS operation.
DEFAULT_FAIL_RETRIES = 8

# The time to wait between retries while performing GCS operation.
DEFAULT_FAIL_WAIT = 2

# Prefix for GCS urls.
GS_PREFIX = 'gs:/'

# Maximum number of cached files per directory.
MAX_CACHED_FILES_PER_DIRECTORY = 15

# https://cloud.google.com/storage/docs/best-practices states that we can
# create/delete 1 bucket about every 2 seconds.
CREATE_BUCKET_DELAY = 4

# GCS blob metadata key for filename.
BLOB_FILENAME_METADATA_KEY = 'filename'

# Thread local globals.
_local = threading.local()

# Urls for web viewer.
OBJECT_URL = 'https://storage.cloud.google.com'
DIRECTORY_URL = 'https://console.cloud.google.com/storage'


class StorageProvider(object):
  """Core storage provider interface."""

  def create_bucket(self, name, object_lifecycle, cors):
    """Create a new bucket."""
    raise NotImplementedError

  def get_bucket(self, name):
    """Get a bucket."""
    raise NotImplementedError

  def list_blobs(self, remote_path, recursive=True):
    """List the blobs under the remote path."""
    raise NotImplementedError

  def copy_file_from(self, remote_path, local_path):
    """Copy file from a remote path to a local path."""
    raise NotImplementedError

  def copy_file_to(self, local_path_or_handle, remote_path, metadata=None):
    """Copy file from a local path to a remote path."""
    raise NotImplementedError

  def copy_blob(self, remote_source, remote_target):
    """Copy a remote file to another remote location."""
    raise NotImplementedError

  def read_data(self, remote_path):
    """Read the data of a remote file."""
    raise NotImplementedError

  def write_data(self, data, remote_path, metadata=None):
    """Write the data of a remote file."""
    raise NotImplementedError

  def get(self, remote_path):
    """Get information about a remote file."""
    raise NotImplementedError

  def delete(self, remote_path):
    """Delete a remote file."""
    raise NotImplementedError


class GcsProvider(StorageProvider):
  """GCS storage provider."""

  def _chunk_size(self):
    if environment.is_running_on_app_engine():
      # To match App Engine URLFetch's request size limit.
      return 10 * 1024 * 1024  # 10 MiB.

    return None

  def create_bucket(self, name, object_lifecycle, cors):
    """Create a new bucket."""
    project_id = utils.get_application_id()
    request_body = {'name': name}
    if object_lifecycle:
      request_body['lifecycle'] = object_lifecycle

    if cors:
      request_body['cors'] = cors

    client = create_discovery_storage_client()
    try:
      client.buckets().insert(project=project_id, body=request_body).execute()
    except HttpError as e:
      logs.log_warn('Failed to create bucket %s: %s' % (name, e))
      raise

    return True

  def get_bucket(self, name):
    """Get a bucket."""
    client = create_discovery_storage_client()
    try:
      return client.buckets().get(bucket=name).execute()
    except HttpError as e:
      if e.resp.status == 404:
        return None

      raise

  def list_blobs(self, remote_path, recursive=True):
    """List the blobs under the remote path."""
    bucket_name, path = get_bucket_name_and_path(remote_path)

    if path and not path.endswith('/'):
      path += '/'

    client = _storage_client()
    bucket = client.bucket(bucket_name)
    properties = {}

    if recursive:
      delimiter = None
    else:
      delimiter = '/'

    iterator = bucket.list_blobs(prefix=path, delimiter=delimiter)
    for blob in iterator:
      properties['bucket'] = bucket_name
      properties['name'] = blob.name
      properties['updated'] = blob.updated
      properties['size'] = blob.size

      yield properties

    if not recursive:
      # When doing delimiter listings, the "directories" will be in `prefixes`.
      for prefix in iterator.prefixes:
        properties['bucket'] = bucket_name
        properties['name'] = prefix
        yield properties

  def copy_file_from(self, remote_path, local_path):
    """Copy file from a remote path to a local path."""
    client = _storage_client()
    bucket_name, path = get_bucket_name_and_path(remote_path)

    try:
      bucket = client.bucket(bucket_name)
      blob = bucket.blob(path, chunk_size=self._chunk_size())
      blob.download_to_filename(local_path)
    except google.cloud.exceptions.GoogleCloudError:
      logs.log_warn('Failed to copy cloud storage file %s to local file %s.' %
                    (remote_path, local_path))
      raise

    return True

  def copy_file_to(self, local_path_or_handle, remote_path, metadata=None):
    """Copy file from a local path to a remote path."""
    client = _storage_client()
    bucket_name, path = get_bucket_name_and_path(remote_path)

    try:
      bucket = client.bucket(bucket_name)
      blob = bucket.blob(path, chunk_size=self._chunk_size())
      if metadata:
        blob.metadata = metadata

      if isinstance(local_path_or_handle, str):
        blob.upload_from_filename(local_path_or_handle)
      else:
        blob.upload_from_file(local_path_or_handle, rewind=True)

    except google.cloud.exceptions.GoogleCloudError:
      logs.log_warn('Failed to copy local file %s to cloud storage file %s.' %
                    (local_path_or_handle, remote_path))
      raise

    return True

  def copy_blob(self, remote_source, remote_target):
    """Copy a remote file to another remote location."""
    source_bucket_name, source_path = get_bucket_name_and_path(remote_source)
    target_bucket_name, target_path = get_bucket_name_and_path(remote_target)

    client = _storage_client()
    try:
      source_bucket = client.bucket(source_bucket_name)
      source_blob = source_bucket.blob(source_path)
      target_bucket = client.bucket(target_bucket_name)
      source_bucket.copy_blob(source_blob, target_bucket, target_path)
    except google.cloud.exceptions.GoogleCloudError:
      logs.log_warn('Failed to copy cloud storage file %s to cloud storage '
                    'file %s.' % (remote_source, remote_target))
      raise

    return True

  def read_data(self, remote_path):
    """Read the data of a remote file."""
    bucket_name, path = get_bucket_name_and_path(remote_path)

    client = _storage_client()
    try:
      bucket = client.bucket(bucket_name)
      blob = bucket.blob(path, chunk_size=self._chunk_size())
      return blob.download_as_string()
    except google.cloud.exceptions.GoogleCloudError as e:
      if e.code == 404:
        return None

      logs.log_warn('Failed to read cloud storage file %s.' % remote_path)
      raise

  def write_data(self, data, remote_path, metadata=None):
    """Write the data of a remote file."""
    client = _storage_client()
    bucket_name, path = get_bucket_name_and_path(remote_path)

    try:
      bucket = client.bucket(bucket_name)
      blob = bucket.blob(path, chunk_size=self._chunk_size())
      if metadata:
        blob.metadata = metadata
      blob.upload_from_string(data)
    except google.cloud.exceptions.GoogleCloudError:
      logs.log_warn('Failed to write cloud storage file %s.' % remote_path)
      raise

    return True

  def get(self, remote_path):
    """Get information about a remote file."""
    client = create_discovery_storage_client()
    bucket, path = get_bucket_name_and_path(remote_path)

    try:
      return client.objects().get(bucket=bucket, object=path).execute()
    except HttpError as e:
      if e.resp.status == 404:
        return None

      raise

  def delete(self, remote_path):
    """Delete a remote file."""
    client = _storage_client()
    bucket_name, path = get_bucket_name_and_path(remote_path)

    try:
      bucket = client.bucket(bucket_name)
      bucket.delete_blob(path)
    except google.cloud.exceptions.GoogleCloudError:
      logs.log_warn('Failed to delete cloud storage file %s.' % remote_path)
      raise

    return True


class FileSystemProvider(StorageProvider):
  """File system backed storage provider."""

  OBJECTS_DIR = 'objects'
  METADATA_DIR = 'metadata'

  def __init__(self, filesystem_dir):
    self.filesystem_dir = os.path.abspath(filesystem_dir)

  def _get_object_properties(self, remote_path):
    """Set local object properties."""
    bucket, path = get_bucket_name_and_path(remote_path)
    fs_path = self.convert_path(remote_path)

    data = {
        'bucket': bucket,
        'name': path,
    }

    if not os.path.isdir(fs_path):
      # These attributes only apply to objects, not directories.
      data.update({
          'updated':
              datetime.datetime.utcfromtimestamp(os.stat(fs_path).st_mtime),
          'size':
              os.path.getsize(fs_path),
          'metadata':
              self._get_metadata(bucket, path),
      })

    return data

  def _get_metadata(self, bucket, path):
    """Get the metadata for a given object."""
    fs_metadata_path = self._fs_path(bucket, path, self.METADATA_DIR)
    if os.path.exists(fs_metadata_path):
      with open(fs_metadata_path) as f:
        return json.load(f)

    return {}

  def _fs_bucket_path(self, bucket):
    """Get the local FS path for the bucket."""
    return os.path.join(self.filesystem_dir, bucket)

  def _fs_objects_dir(self, bucket):
    """Get the local FS path for objects in the bucket."""
    return os.path.join(self._fs_bucket_path(bucket), self.OBJECTS_DIR)

  def _fs_path(self, bucket, path, directory):
    """Get the local object/metadata FS path."""
    return os.path.join(self._fs_bucket_path(bucket), directory, path)

  def _write_metadata(self, remote_path, metadata):
    """Write metadata."""
    if not metadata:
      return

    fs_metadata_path = self.convert_path_for_write(remote_path,
                                                   self.METADATA_DIR)
    with open(fs_metadata_path, 'w') as f:
      json.dump(metadata, f)

  def convert_path(self, remote_path, directory=OBJECTS_DIR):
    """Get the local FS path for the remote path."""
    bucket, path = get_bucket_name_and_path(remote_path)
    return self._fs_path(bucket, path, directory)

  def convert_path_for_write(self, remote_path, directory=OBJECTS_DIR):
    """Get the local FS path for writing to the remote path. Creates any
    intermediate directories if necessary (except for the parent bucket
    directory)."""
    bucket, path = get_bucket_name_and_path(remote_path)
    if not os.path.exists(self._fs_bucket_path(bucket)):
      raise RuntimeError(
          'Bucket {bucket} does not exist.'.format(bucket=bucket))

    fs_path = self._fs_path(bucket, path, directory)
    shell.create_directory(os.path.dirname(fs_path), create_intermediates=True)

    return fs_path

  def create_bucket(self, name, object_lifecycle, cors):
    """Create a new bucket."""
    bucket_path = self._fs_bucket_path(name)
    if os.path.exists(bucket_path):
      return False

    os.makedirs(bucket_path)
    return True

  def get_bucket(self, name):
    """Get a bucket."""
    bucket_path = self._fs_bucket_path(name)
    if not os.path.exists(bucket_path):
      return None

    return {
        'name': name,
    }

  def _list_files_recursive(self, fs_path):
    """List files recursively."""
    for root, _, filenames in shell.walk(fs_path):
      for filename in filenames:
        yield os.path.join(root, filename)

  def _list_files_nonrecursive(self, fs_path):
    """List files non-recursively."""
    for filename in os.listdir(fs_path):
      yield os.path.join(fs_path, filename)

  def list_blobs(self, remote_path, recursive=True):
    """List the blobs under the remote path."""
    bucket, _ = get_bucket_name_and_path(remote_path)
    fs_path = self.convert_path(remote_path)

    if recursive:
      file_paths = self._list_files_recursive(fs_path)
    else:
      file_paths = self._list_files_nonrecursive(fs_path)

    for fs_path in file_paths:
      path = os.path.relpath(fs_path, self._fs_objects_dir(bucket))

      yield self._get_object_properties(
          get_cloud_storage_file_path(bucket, path))

  def copy_file_from(self, remote_path, local_path):
    """Copy file from a remote path to a local path."""
    fs_path = self.convert_path(remote_path)
    return shell.copy_file(fs_path, local_path)

  def copy_file_to(self, local_path_or_handle, remote_path, metadata=None):
    """Copy file from a local path to a remote path."""
    fs_path = self.convert_path_for_write(remote_path)

    if isinstance(local_path_or_handle, str):
      if not shell.copy_file(local_path_or_handle, fs_path):
        return False
    else:
      with open(fs_path, 'wb') as f:
        shutil.copyfileobj(local_path_or_handle, f)

    self._write_metadata(remote_path, metadata)
    return True

  def copy_blob(self, remote_source, remote_target):
    """Copy a remote file to another remote location."""
    fs_source_path = self.convert_path(remote_source)
    fs_target_path = self.convert_path_for_write(remote_target)
    return shell.copy_file(fs_source_path, fs_target_path)

  def read_data(self, remote_path):
    """Read the data of a remote file."""
    fs_path = self.convert_path(remote_path)
    if not os.path.exists(fs_path):
      return None

    with open(fs_path, 'rb') as f:
      return f.read()

  def write_data(self, data, remote_path, metadata=None):
    """Write the data of a remote file."""
    fs_path = self.convert_path_for_write(remote_path)
    if isinstance(data, str):
      data = data.encode()

    with open(fs_path, 'wb') as f:
      f.write(data)

    self._write_metadata(remote_path, metadata)
    return True

  def get(self, remote_path):
    """Get information about a remote file."""
    fs_path = self.convert_path(remote_path)
    if not os.path.exists(fs_path):
      return None

    return self._get_object_properties(remote_path)

  def delete(self, remote_path):
    """Delete a remote file."""
    fs_path = self.convert_path(remote_path)
    shell.remove_file(fs_path)

    fs_metadata_path = self.convert_path(remote_path, self.METADATA_DIR)
    shell.remove_file(fs_metadata_path)
    return True


class GcsBlobInfo(object):
  """GCS blob info."""

  def __init__(self,
               bucket,
               object_path,
               filename=None,
               size=None,
               legacy_key=None):
    self.bucket = bucket
    self.object_path = object_path

    if filename is not None and size is not None:
      self.filename = filename
      self.size = size
    else:
      gcs_object = get(get_cloud_storage_file_path(bucket, object_path))

      self.filename = gcs_object['metadata'].get(BLOB_FILENAME_METADATA_KEY)
      self.size = int(gcs_object['size'])

    self.legacy_key = legacy_key

  def key(self):
    if self.legacy_key:
      return self.legacy_key

    return self.object_path

  @property
  def gcs_path(self):
    return '/%s/%s' % (self.bucket, self.object_path)

  @staticmethod
  def from_key(key):
    try:
      return GcsBlobInfo(blobs_bucket(), key)
    except Exception:
      logs.log_error('Failed to get blob from key %s.' % key)
      return None

  @staticmethod
  def from_legacy_blob_info(blob_info):
    bucket, path = get_bucket_name_and_path(blob_info.gs_object_name)
    return GcsBlobInfo(bucket, path, blob_info.filename, blob_info.size,
                       blob_info.key.id())


def _provider():
  """Get the current storage provider."""
  local_buckets_path = environment.get_value('LOCAL_GCS_BUCKETS_PATH')
  if local_buckets_path:
    return FileSystemProvider(local_buckets_path)

  return GcsProvider()


def _create_storage_client_new():
  """Create a storage client."""
  creds, project = credentials.get_default()
  if not project:
    project = utils.get_application_id()

  return gcs.Client(project=project, credentials=creds)


def _storage_client():
  """Get the storage client, creating it if it does not exist."""
  if hasattr(_local, 'client'):
    return _local.client

  _local.client = _create_storage_client_new()
  return _local.client


def get_bucket_name_and_path(cloud_storage_file_path):
  """Return bucket name and path given a full cloud storage path."""
  filtered_path = utils.strip_from_left(cloud_storage_file_path, GS_PREFIX)
  _, bucket_name_and_path = filtered_path.split('/', 1)

  if '/' in bucket_name_and_path:
    bucket_name, path = bucket_name_and_path.split('/', 1)
  else:
    bucket_name = bucket_name_and_path
    path = ''

  return bucket_name, path


def get_cloud_storage_file_path(bucket, path):
  """Get the full GCS file path."""
  return GS_PREFIX + '/' + bucket + '/' + path


def _get_error_reason(http_error):
  """Get error reason from googleapiclient.errors.HttpError."""
  try:
    data = json.loads(http_error.content.decode('utf-8'))
    return data['error']['message']
  except (ValueError, KeyError):
    logs.log_error('Failed to decode error content: %s' % http_error.content)

  return None


@environment.local_noop
def add_single_bucket_iam(storage, iam_policy, role, bucket_name, member):
  """Attempt to add a single bucket IAM. Returns the modified iam policy, or
  None on failure."""
  binding = get_bucket_iam_binding(iam_policy, role)
  binding['members'].append(member)

  result = set_bucket_iam_policy(storage, bucket_name, iam_policy)

  binding['members'].pop()
  return result


@environment.local_noop
def get_bucket_iam_binding(iam_policy, role):
  """Get the binding matching a role, or None."""
  return next((
      binding for binding in iam_policy['bindings'] if binding['role'] == role),
              None)


@environment.local_noop
def get_or_create_bucket_iam_binding(iam_policy, role):
  """Get or create the binding matching a role."""
  binding = get_bucket_iam_binding(iam_policy, role)
  if not binding:
    binding = {'role': role, 'members': []}
    iam_policy['bindings'].append(binding)

  return binding


@environment.local_noop
def remove_bucket_iam_binding(iam_policy, role):
  """Remove existing binding matching the role."""
  iam_policy['bindings'] = [
      binding for binding in iam_policy['bindings'] if binding['role'] != role
  ]


@environment.local_noop
def get_bucket_iam_policy(storage, bucket_name):
  """Get bucket IAM policy."""
  try:
    iam_policy = storage.buckets().getIamPolicy(bucket=bucket_name).execute()
  except HttpError as e:
    logs.log_error('Failed to get IAM policies for %s: %s' % (bucket_name, e))
    return None

  return iam_policy


@environment.local_noop
def set_bucket_iam_policy(client, bucket_name, iam_policy):
  """Set bucket IAM policy."""
  filtered_iam_policy = copy.deepcopy(iam_policy)

  # Bindings returned by getIamPolicy can have duplicates. Remove them or
  # otherwise, setIamPolicy operation fails.
  for binding in filtered_iam_policy['bindings']:
    binding['members'] = sorted(list(set(binding['members'])))

  # Filtering members can cause a binding to have no members. Remove binding
  # or otherwise, setIamPolicy operation fails.
  filtered_iam_policy['bindings'] = [
      b for b in filtered_iam_policy['bindings'] if b['members']
  ]

  try:
    return client.buckets().setIamPolicy(
        bucket=bucket_name, body=filtered_iam_policy).execute()
  except HttpError as e:
    error_reason = _get_error_reason(e)
    if error_reason == 'Invalid argument':
      # Expected error for non-Google emails or groups. Warn about these.
      logs.log_warn('Invalid Google email or group being added to bucket %s.' %
                    bucket_name)
    elif error_reason and 'is of type "group"' in error_reason:
      logs.log_warn('Failed to set IAM policy for %s bucket for a group: %s.' %
                    (bucket_name, error_reason))
    else:
      logs.log_error('Failed to set IAM policies for bucket %s.' % bucket_name)

  return None


def create_bucket_if_needed(bucket_name, object_lifecycle=None, cors=None):
  """Creates a GCS bucket."""
  provider = _provider()
  if provider.get_bucket(bucket_name):
    return True

  if not provider.create_bucket(bucket_name, object_lifecycle, cors):
    return False

  time.sleep(CREATE_BUCKET_DELAY)
  return True


@environment.local_noop
def create_discovery_storage_client():
  """Create a storage client using discovery APIs."""
  return build('storage', 'v1', cache_discovery=False)


def generate_life_cycle_config(action, age=None, num_newer_versions=None):
  """Generate GCS lifecycle management config.

  For the reference, see https://cloud.google.com/storage/docs/lifecycle and
  https://cloud.google.com/storage/docs/managing-lifecycles.
  """
  rule = {}
  rule['action'] = {'type': action}
  rule['condition'] = {}
  if age is not None:
    rule['condition']['age'] = age
  if num_newer_versions is not None:
    rule['condition']['numNewerVersions'] = num_newer_versions

  config = {'rule': [rule]}
  return config


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.copy_file_from',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def copy_file_from(cloud_storage_file_path, local_file_path, use_cache=False):
  """Saves a cloud storage file locally."""
  if use_cache and get_file_from_cache_if_exists(local_file_path):
    logs.log('Copied file %s from local cache.' % cloud_storage_file_path)
    return True

  if not _provider().copy_file_from(cloud_storage_file_path, local_file_path):
    return False

  if use_cache:
    store_file_in_cache(local_file_path)

  return True


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.copy_file_to',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def copy_file_to(local_file_path_or_handle,
                 cloud_storage_file_path,
                 metadata=None):
  """Copy local file to a cloud storage path."""
  if (isinstance(local_file_path_or_handle, str) and
      not os.path.exists(local_file_path_or_handle)):
    logs.log_error('Local file %s not found.' % local_file_path_or_handle)
    return False

  return _provider().copy_file_to(
      local_file_path_or_handle, cloud_storage_file_path, metadata=metadata)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.copy_blob',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def copy_blob(cloud_storage_source_path, cloud_storage_target_path):
  """Copy two blobs on GCS 'in the cloud' without touching local disk."""
  return _provider().copy_blob(cloud_storage_source_path,
                               cloud_storage_target_path)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.delete',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def delete(cloud_storage_file_path):
  """Delete a cloud storage file given its path."""
  return _provider().delete(cloud_storage_file_path)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.exists')
def exists(cloud_storage_file_path, ignore_errors=False):
  """Return whether if a cloud storage file exists."""
  try:
    return bool(_provider().get(cloud_storage_file_path))
  except HttpError:
    if not ignore_errors:
      logs.log_error('Failed when trying to find cloud storage file %s.' %
                     cloud_storage_file_path)

    return False


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.last_updated',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def last_updated(cloud_storage_file_path):
  """Return last updated value by parsing stats for all blobs under a cloud
  storage path."""
  last_update = None
  for blob in _provider().list_blobs(cloud_storage_file_path):
    if not last_update or blob['updated'] > last_update:
      last_update = blob['updated']
  if last_update:
    # Remove UTC tzinfo to make these comparable.
    last_update = last_update.replace(tzinfo=None)
  return last_update


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.read_data',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def read_data(cloud_storage_file_path):
  """Return content of a cloud storage file."""
  return _provider().read_data(cloud_storage_file_path)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.write_data',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def write_data(data, cloud_storage_file_path, metadata=None):
  """Return content of a cloud storage file."""
  return _provider().write_data(
      data, cloud_storage_file_path, metadata=metadata)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.get_blobs',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def get_blobs(cloud_storage_path, recursive=True):
  """Return blobs under the given cloud storage path."""
  for blob in _provider().list_blobs(cloud_storage_path, recursive=recursive):
    yield blob


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.list_blobs',
    exception_type=google.cloud.exceptions.GoogleCloudError)
def list_blobs(cloud_storage_path, recursive=True):
  """Return blob names under the given cloud storage path."""
  for blob in _provider().list_blobs(cloud_storage_path, recursive=recursive):
    yield blob['name']


def get_download_file_size(cloud_storage_file_path,
                           file_path=None,
                           use_cache=False):
  """Get the download file size of the bucket path."""
  if use_cache and file_path:
    size_from_cache = get_file_size_from_cache_if_exists(file_path)
    if size_from_cache is not None:
      return size_from_cache

  return get_object_size(cloud_storage_file_path)


@utils.timeout(CACHE_TIMEOUT)
def get_file_from_cache_if_exists(file_path,
                                  update_modification_time_on_access=True):
  """Get file from nfs cache if available."""
  cache_file_path = get_cache_file_path(file_path)
  if not cache_file_path or not file_exists_in_cache(cache_file_path):
    # If the file does not exist in cache, bail out.
    return False

  # Fetch cache file size before starting the actual copy.
  cache_file_size = get_cache_file_size_from_metadata(cache_file_path)

  # Copy file from cache to local.
  if not shell.copy_file(cache_file_path, file_path):
    return False

  # Update timestamp to later help with eviction of old files.
  if update_modification_time_on_access:
    update_access_and_modification_timestamp(cache_file_path)

  # Return success or failure based on existence of local file and size
  # comparison.
  return (os.path.exists(file_path) and
          os.path.getsize(file_path) == cache_file_size)


@utils.timeout(CACHE_TIMEOUT)
def get_file_size_from_cache_if_exists(file_path):
  """Get file size from nfs cache if available."""
  cache_file_path = get_cache_file_path(file_path)
  if not cache_file_path or not file_exists_in_cache(cache_file_path):
    # If the file does not exist in cache, bail out.
    return None

  return get_cache_file_size_from_metadata(cache_file_path)


def get_cache_file_path(file_path):
  """Return cache file path given a local file path."""
  if not environment.get_value('NFS_ROOT'):
    return None

  return os.path.join(
      environment.get_value('NFS_ROOT'), CACHE_DIRNAME,
      utils.get_directory_hash_for_path(file_path), os.path.basename(file_path))


def get_cache_file_metadata_path(cache_file_path):
  """Return metadata file path for a cache file."""
  return '%s%s' % (cache_file_path, CACHE_METADATA_FILE_EXTENSION)


def get_cache_file_size_from_metadata(cache_file_path):
  """Return cache file size from metadata file."""
  cache_file_metadata_path = get_cache_file_metadata_path(cache_file_path)
  metadata_content = utils.read_data_from_file(
      cache_file_metadata_path, eval_data=True)

  if not metadata_content or 'size' not in metadata_content:
    return None

  return metadata_content['size']


def write_cache_file_metadata(cache_file_path, file_path):
  """Write cache file metadata."""
  cache_file_metadata_path = get_cache_file_metadata_path(cache_file_path)
  utils.write_data_to_file({
      'size': os.path.getsize(file_path)
  }, cache_file_metadata_path)


def remove_cache_file_and_metadata(cache_file_path):
  """Removes cache file and its metadata."""
  logs.log('Removing cache file %s and its metadata.' % cache_file_path)
  shell.remove_file(get_cache_file_metadata_path(cache_file_path))
  shell.remove_file(cache_file_path)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.'
    'update_access_and_modification_timestamp')
def update_access_and_modification_timestamp(file_path):
  os.utime(file_path, None)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.file_exists_in_cache')
def file_exists_in_cache(cache_file_path):
  """Returns if the file exists in cache."""
  cache_file_metadata_path = get_cache_file_metadata_path(cache_file_path)
  if not os.path.exists(cache_file_metadata_path):
    return False

  if not os.path.exists(cache_file_path):
    return False

  actual_cache_file_size = os.path.getsize(cache_file_path)
  expected_cache_file_size = get_cache_file_size_from_metadata(cache_file_path)
  return actual_cache_file_size == expected_cache_file_size


@utils.timeout(CACHE_TIMEOUT)
def store_file_in_cache(file_path,
                        cached_files_per_directory_limit=True,
                        force_update=False):
  """Get file from nfs cache if available."""
  if not os.path.exists(file_path):
    logs.log_error(
        'Local file %s does not exist, nothing to store in cache.' % file_path)
    return

  if os.path.getsize(file_path) > CACHE_SIZE_LIMIT:
    logs.log('File %s is too large to store in cache, skipping.' % file_path)
    return

  nfs_root = environment.get_value('NFS_ROOT')
  if not nfs_root:
    # No NFS, nothing to store in cache.
    return

  # If NFS server is not available due to heavy load, skip storage operation
  # altogether as we would fail to store file.
  if not os.path.exists(os.path.join(nfs_root, '.')):  # Use . to iterate mount.
    logs.log_warn('Cache %s not available.' % nfs_root)
    return

  cache_file_path = get_cache_file_path(file_path)
  cache_directory = os.path.dirname(cache_file_path)
  filename = os.path.basename(file_path)

  if not os.path.exists(cache_directory):
    if not shell.create_directory(cache_directory, create_intermediates=True):
      logs.log_error('Failed to create cache directory %s.' % cache_directory)
      return

  # Check if the file already exists in cache.
  if file_exists_in_cache(cache_file_path):
    if not force_update:
      return

    # If we are forcing update, we need to remove current cached file and its
    # metadata.
    remove_cache_file_and_metadata(cache_file_path)

  # Delete old cached files beyond our maximum storage limit.
  if cached_files_per_directory_limit:
    # Get a list of cached files.
    cached_files_list = []
    for cached_filename in os.listdir(cache_directory):
      if cached_filename.endswith(CACHE_METADATA_FILE_EXTENSION):
        continue
      cached_file_path = os.path.join(cache_directory, cached_filename)
      cached_files_list.append(cached_file_path)

    mtime = lambda f: os.stat(f).st_mtime
    last_used_cached_files_list = list(
        sorted(cached_files_list, key=mtime, reverse=True))
    for cached_file_path in (
        last_used_cached_files_list[MAX_CACHED_FILES_PER_DIRECTORY - 1:]):
      remove_cache_file_and_metadata(cached_file_path)

  # Start storing the actual file in cache now.
  logs.log('Started storing file %s into cache.' % filename)

  # Fetch lock to store this file. Try only once since if any other bot has
  # started to store it, we don't need to do it ourselves. Just bail out.
  lock_name = 'store:cache_file:%s' % utils.string_hash(cache_file_path)
  if not locks.acquire_lock(
      lock_name, max_hold_seconds=CACHE_LOCK_TIMEOUT, retries=1, by_zone=True):
    logs.log_warn(
        'Unable to fetch lock to update cache file %s, skipping.' % filename)
    return

  # Check if another bot already updated it.
  if file_exists_in_cache(cache_file_path):
    locks.release_lock(lock_name, by_zone=True)
    return

  shell.copy_file(file_path, cache_file_path)
  write_cache_file_metadata(cache_file_path, file_path)
  time.sleep(CACHE_COPY_WAIT_TIME)
  error_occurred = not file_exists_in_cache(cache_file_path)
  locks.release_lock(lock_name, by_zone=True)

  if error_occurred:
    logs.log_error('Failed to store file %s into cache.' % filename)
  else:
    logs.log('Completed storing file %s into cache.' % filename)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.get')
def get(cloud_storage_file_path):
  """Get GCS object data."""
  return _provider().get(cloud_storage_file_path)


@environment.local_noop
@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.get_acl')
def get_acl(cloud_storage_file_path, entity):
  """Get the access control for a file."""
  client = create_discovery_storage_client()
  bucket, path = get_bucket_name_and_path(cloud_storage_file_path)

  try:
    return client.objectAccessControls().get(
        bucket=bucket, object=path, entity=entity).execute()
  except HttpError as e:
    if e.resp.status == 404:
      return None

    raise


@environment.local_noop
@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='google_cloud_utils.storage.set_acl')
def set_acl(cloud_storage_file_path, entity, role='READER'):
  """Set the access control for a file."""
  client = create_discovery_storage_client()
  bucket, path = get_bucket_name_and_path(cloud_storage_file_path)

  try:
    return client.objectAccessControls().insert(
        bucket=bucket, object=path, body={
            'entity': entity,
            'role': role
        }).execute()
  except HttpError as e:
    if e.resp.status == 404:
      return None

    raise


def get_object_size(cloud_storage_file_path):
  """Get the metadata for a file."""
  gcs_object = get(cloud_storage_file_path)
  if not gcs_object:
    return gcs_object

  return int(gcs_object['size'])


def blobs_bucket():
  """Get the blobs bucket name."""
  # Allow tests to override blobs bucket name safely.
  test_blobs_bucket = environment.get_value('TEST_BLOBS_BUCKET')
  if test_blobs_bucket:
    return test_blobs_bucket

  assert not environment.get_value('PY_UNITTESTS')
  return local_config.ProjectConfig().get('blobs.bucket')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for adb functions."""

import os
import tempfile

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.platforms.android import constants
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import android_helpers
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class CopyLocalDirectoryToRemoteTest(android_helpers.AndroidTest):
  """Tests copy_local_directory_to_remote."""

  def setUp(self):
    super(CopyLocalDirectoryToRemoteTest, self).setUp()

    # Clear and create temporary directory on device.
    self.device_temp_dir = constants.DEVICE_TMP_DIR
    adb.remove_directory(self.device_temp_dir, recreate=True)

    # Create local temp directory.
    self.local_temp_dir = tempfile.mkdtemp()

  def tearDown(self):
    adb.remove_directory(self.device_temp_dir)
    shell.remove_directory(self.local_temp_dir)

  def test(self):
    """Tests copy_local_directory_to_remote."""
    utils.write_data_to_file('a', os.path.join(self.local_temp_dir, 'a'))
    shell.create_directory(os.path.join(self.local_temp_dir, 'b'))
    utils.write_data_to_file('c', os.path.join(self.local_temp_dir, 'b', 'c'))

    adb.copy_local_directory_to_remote(self.local_temp_dir,
                                       self.device_temp_dir)

    self.assertTrue(adb.file_exists(os.path.join(self.device_temp_dir, 'a')))
    self.assertFalse(
        adb.directory_exists(os.path.join(self.device_temp_dir, 'a')))
    self.assertEqual(
        adb.get_file_size(os.path.join(self.device_temp_dir, 'a')), 1)

    self.assertTrue(
        adb.directory_exists(os.path.join(self.device_temp_dir, 'b')))
    self.assertFalse(adb.file_exists(os.path.join(self.device_temp_dir, 'b')))

    self.assertTrue(
        adb.file_exists(os.path.join(self.device_temp_dir, 'b', 'c')))
    self.assertFalse(
        adb.directory_exists(os.path.join(self.device_temp_dir, 'b', 'c')))
    self.assertEqual(
        adb.get_file_size(os.path.join(self.device_temp_dir, 'b', 'c')), 1)


class CopyRemoteDirectoryToLocalTest(android_helpers.AndroidTest):
  """Tests copy_remote_directory_to_local."""

  def setUp(self):
    super(CopyRemoteDirectoryToLocalTest, self).setUp()

    # Clear and create temporary directory on device.
    self.device_temp_dir = constants.DEVICE_TMP_DIR
    adb.remove_directory(self.device_temp_dir, recreate=True)

    # Create local temp directory.
    self.local_temp_dir = tempfile.mkdtemp()

  def tearDown(self):
    adb.remove_directory(self.device_temp_dir)
    shell.remove_directory(self.local_temp_dir)

  def test(self):
    """Tests copy_remote_directory_to_local."""
    adb.write_data_to_file('a', os.path.join(self.device_temp_dir, 'a'))
    adb.create_directory_if_needed(os.path.join(self.device_temp_dir, 'b'))
    adb.write_data_to_file('c', os.path.join(self.device_temp_dir, 'b', 'c'))

    adb.copy_remote_directory_to_local(self.device_temp_dir,
                                       self.local_temp_dir)

    self.assertTrue(os.path.exists(os.path.join(self.local_temp_dir, 'a')))
    self.assertTrue(os.path.isfile(os.path.join(self.local_temp_dir, 'a')))
    self.assertEqual(os.path.getsize(os.path.join(self.local_temp_dir, 'a')), 1)

    self.assertTrue(os.path.exists(os.path.join(self.local_temp_dir, 'b')))
    self.assertTrue(os.path.isdir(os.path.join(self.local_temp_dir, 'b')))

    self.assertTrue(os.path.exists(os.path.join(self.local_temp_dir, 'b', 'c')))
    self.assertTrue(os.path.isfile(os.path.join(self.local_temp_dir, 'b', 'c')))
    self.assertEqual(
        os.path.getsize(os.path.join(self.local_temp_dir, 'b', 'c')), 1)


class ReadDataFromFileTest(android_helpers.AndroidTest):
  """Tests read_data_from_file."""

  def setUp(self):
    super(ReadDataFromFileTest, self).setUp()

    # Clear and create temporary directory on device.
    self.device_temp_dir = constants.DEVICE_TMP_DIR
    adb.remove_directory(self.device_temp_dir, recreate=True)

  def tearDown(self):
    adb.remove_directory(self.device_temp_dir)

  def test_non_existent_file(self):
    """Test that we return None when file does not exist."""
    non_existent_file_path = os.path.join(self.device_temp_dir, 'non-existent')
    self.assertEqual(adb.read_data_from_file(non_existent_file_path), None)

  def test_regular_file(self):
    """Test that file data is read from regular file."""
    test_file_path = os.path.join(self.device_temp_dir, 'a')
    adb.write_data_to_file('a' * 5000, test_file_path)
    self.assertEqual(adb.read_data_from_file(test_file_path), 'a' * 5000)


class WriteDataToFileTest(android_helpers.AndroidTest):
  """Tests write_data_to_file."""

  def setUp(self):
    super(WriteDataToFileTest, self).setUp()

    # Clear and create temporary directory on device.
    self.device_temp_dir = constants.DEVICE_TMP_DIR
    adb.remove_directory(self.device_temp_dir, recreate=True)

  def tearDown(self):
    adb.remove_directory(self.device_temp_dir)

  def test_regular_file(self):
    """Test that file data is written."""
    test_file_path = os.path.join(self.device_temp_dir, 'a')
    adb.write_data_to_file('a' * 5000, test_file_path)
    self.assertEqual(adb.read_data_from_file(test_file_path), 'a' * 5000)


class GetFileSizeTest(android_helpers.AndroidTest):
  """Tests get_file_size."""

  def test_nonexistent_file(self):
    """Tests that file size is None for a non-existent file."""
    self.assertFalse(adb.file_exists('/sdcard/non-existent'))
    self.assertIsNone(adb.get_file_size('/sdcard/non-existent'))

  def test_regular_file(self):
    """Test that file size is properly calculated for regular files."""
    adb.write_data_to_file('abcd', '/sdcard/file1')
    self.assertTrue(adb.file_exists('/sdcard/file1'))
    self.assertEqual(adb.get_file_size('/sdcard/file1'), 4)


class GetFileChecksumTest(android_helpers.AndroidTest):
  """Tests get_file_checksum."""

  def test_nonexistent_file(self):
    """Tests that checksum is None for a non-existent file."""
    self.assertFalse(adb.file_exists('/sdcard/non-existent'))
    self.assertIsNone(adb.get_file_checksum('/sdcard/non-existent'))

  def test_regular_file(self):
    """Test that checksum is properly calculated for regular files."""
    adb.write_data_to_file('abcd', '/sdcard/file1')
    self.assertTrue(adb.file_exists('/sdcard/file1'))
    self.assertEqual(
        adb.get_file_checksum('/sdcard/file1'),
        'e2fc714c4727ee9395f324cd2e7f331f')


class GetPSOutputTest(android_helpers.AndroidTest):
  """Tests get_ps_output."""

  def test(self):
    """Test that ps output is parseable for pid and ppids."""
    ps_output = adb.get_ps_output()
    ps_output_lines = ps_output.splitlines()[1:]  # Skip column line.
    for line in ps_output_lines:
      values = line.split()
      self.assertTrue(values[1].isdigit())  # PID.
      self.assertTrue(values[2].isdigit())  # PPID


class WaitForDeviceTest(android_helpers.AndroidTest):
  """Tests for wait_for_device."""

  def test_state_correct_after_wait(self):
    """Ensure that the function works correctly when a device is connected."""
    adb.wait_for_device()
    self.assertEqual(adb.get_device_state(), 'device')


class ResetUsbTest(android_helpers.AndroidTest):
  """Tests for reset_usb."""

  def setUp(self):
    super(ResetUsbTest, self).setUp()
    test_helpers.patch(self, ['fcntl.ioctl'])

  def test_with_device(self):
    """Tests reset_usb with a connected device if available."""
    self.assertTrue(adb.reset_usb())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for untrusted runner."""

import functools

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.system import environment


class HostException(SystemExit):
  """Unrecoverable Exception."""


def untrusted_noop(return_value=None):
  """Return a decorator that turns functions into no-ops if the bot is
  untrusted."""

  def decorator(func):
    """Decorator function."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
      if environment.is_untrusted_worker():
        return return_value

      return func(*args, **kwargs)

    return wrapped

  return decorator


def internal_network_domain():
  """Return the internal network domain."""
  return '.c.%s.internal' % utils.get_application_id()


def platform_name(project, platform):
  """"Get the untrusted platform name."""
  return project.upper() + '_' + platform.upper()


def queue_name(project, platform):
  """Get the untrusted queue name for the project and platform."""
  return tasks.queue_for_platform(platform_name(project, platform))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helps with remote command-line tasks (e.g. read logs, stage zip)."""

import inspect
import paramiko
import six

from fabric import api

from local.remote.handlers import android_chrome_lab
from local.remote.handlers import linux
from local.remote.handlers import mac
from local.remote.handlers import windows

paramiko.util.log_to_file('paramiko.log')
api.env.connection_attempts = 3
api.env.timeout = 3


def _get_handler_ctor(args):
  """Get a Handler class given arguments."""
  if 'linux' in args.instance_name:
    return linux.Handler
  if 'windows' in args.instance_name:
    return windows.Handler
  if 'golo' in args.instance_name:
    return mac.Handler
  if 'android-build' in args.instance_name:
    return android_chrome_lab.Handler
  raise NotImplementedError('Unsupported platform.')


def _args_to_dict(args, method):
  """Convert args to dict that is compatible with the method's argument."""
  arg_names = inspect.getfullargspec(method).args[1:]
  args_dict = {
      k: v
      for k, v in six.iteritems(vars(args))
      if k in arg_names and v is not None
  }
  return args_dict


def execute(args):
  """Run command-line tasks on a remote bot."""
  handler_ctor = _get_handler_ctor(args)
  handler = handler_ctor(**_args_to_dict(args, handler_ctor.__init__))

  method = getattr(handler, args.remote)
  method(**_args_to_dict(args, method))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the fuzzer-coverage cron task."""

import datetime
import unittest

from google.cloud import ndb

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import fuzzer_coverage

INTEGRATION_TEST_BUCKET = 'clusterfuzz-test-data'


@test_utils.integration
@test_utils.with_cloud_emulators('datastore')
class HandlerTest(unittest.TestCase):
  """Tests for the fuzzer coverage crom task Handler."""

  def setUp(self):
    test_helpers.patch(
        self, ['clusterfuzz._internal.base.utils.default_project_name'])

    # It's important to use chromium here to test they datastore key generation
    # logic which uses project qualified fuzz target names.
    self.mock.default_project_name.return_value = 'chromium'

  def assertCoverageInformation(
      self,
      actual,
      expected,
  ):
    """Assert that the actual entity is equal to the expected one."""
    self.assertEqual(actual.date, expected.date)
    self.assertEqual(actual.fuzzer, expected.fuzzer)
    self.assertEqual(actual.functions_covered, expected.functions_covered)
    self.assertEqual(actual.functions_total, expected.functions_total)
    self.assertEqual(actual.edges_covered, expected.edges_covered)
    self.assertEqual(actual.edges_total, expected.edges_total)
    self.assertEqual(actual.corpus_size_units, expected.corpus_size_units)
    self.assertEqual(actual.corpus_size_bytes, expected.corpus_size_bytes)
    self.assertEqual(actual.corpus_location, expected.corpus_location)
    self.assertEqual(actual.corpus_backup_location,
                     expected.corpus_backup_location)
    self.assertEqual(actual.quarantine_size_units,
                     expected.quarantine_size_units)
    self.assertEqual(actual.quarantine_size_bytes,
                     expected.quarantine_size_bytes)
    self.assertEqual(actual.quarantine_location, expected.quarantine_location)
    self.assertEqual(actual.html_report_url, expected.html_report_url)

  def test_fuzzer_coverage(self):
    """Test fuzzer coverage cron implementation."""
    # An old CoverageInformation for a fuzzer that should NOT be overwritten.
    cov_info_old = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 1),
        fuzzer='boringssl_privkey',
        functions_covered=123,
        functions_total=555,
        edges_covered=1337,
        edges_total=31337,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/boringssl/'
            'reports/20180905/linux/index.html'))
    cov_info_old.put()

    # A recent CoverageInformation for a fuzzer that should be overwritten.
    cov_info_recent = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='base64_decode_fuzzer',
        functions_covered=1,
        functions_total=5,
        edges_covered=3,
        edges_total=20,
        html_report_url='intentionally junk URL that must be overwritten')
    cov_info_recent.put()

    # A recent CoverageInformation for a project that should be overwritten.
    cov_info_project = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='zlib',
        functions_covered=1,
        functions_total=2,
        edges_covered=3,
        edges_total=4,
        html_report_url='intentionally junk URL that must be overwritten')
    cov_info_project.put()

    fuzzer_coverage.collect_fuzzer_coverage(INTEGRATION_TEST_BUCKET)
    query = data_types.CoverageInformation.query()

    entities = {}
    for cov_info in query.fetch():
      entities[cov_info.key] = cov_info

    # Assert and delete entities one by one to make sure we verify each of them.
    key = ndb.Key('CoverageInformation', 'boringssl_bn_div-20180905')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 5),
        fuzzer='boringssl_bn_div',
        functions_covered=82,
        functions_total=1079,
        edges_covered=1059,
        edges_total=12384,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/boringssl/'
            'reports/20180905/linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    # This is the "old" entity that should not be updated (|cov_info_old|).
    key = ndb.Key('CoverageInformation', 'boringssl_privkey-20180901')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 1),
        fuzzer='boringssl_privkey',
        functions_covered=123,
        functions_total=555,
        edges_covered=1337,
        edges_total=31337,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/boringssl/'
            'reports/20180905/linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    key = ndb.Key('CoverageInformation', 'boringssl_privkey-20180905')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 5),
        fuzzer='boringssl_privkey',
        functions_covered=374,
        functions_total=1510,
        edges_covered=3535,
        edges_total=16926,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/boringssl/'
            'reports/20180905/linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    # Entity for boringssl project, not for a single fuzz target.
    key = ndb.Key('CoverageInformation', 'boringssl-20180905')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 5),
        fuzzer='boringssl',
        functions_covered=1872,
        functions_total=4137,
        edges_covered=21303,
        edges_total=51251,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/boringssl/'
            'reports/20180905/linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    # The "recent" entity that should be updated (|cov_info_recent|).
    key = ndb.Key('CoverageInformation', 'base64_decode_fuzzer-20180907')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='base64_decode_fuzzer',
        functions_covered=252,
        functions_total=5646,
        edges_covered=1111,
        edges_total=38748,
        html_report_url=(
            'https://chromium-coverage.appspot.com/reports/589371_fuzzers_only/'
            'linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    key = ndb.Key('CoverageInformation', 'zucchini_raw_gen_fuzzer-20180907')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='zucchini_raw_gen_fuzzer',
        functions_covered=440,
        functions_total=6439,
        edges_covered=1791,
        edges_total=45121,
        html_report_url=(
            'https://chromium-coverage.appspot.com/reports/589371_fuzzers_only/'
            'linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    # Entity for chromium project.
    key = ndb.Key('CoverageInformation', 'chromium-20180907')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='chromium',
        functions_covered=79960,
        functions_total=467023,
        edges_covered=682323,
        edges_total=3953229,
        html_report_url=(
            'https://chromium-coverage.appspot.com/reports/589371_fuzzers_only/'
            'linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    key = ndb.Key('CoverageInformation', 'zlib_uncompress_fuzzer-20180907')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='zlib_uncompress_fuzzer',
        functions_covered=19,
        functions_total=47,
        edges_covered=987,
        edges_total=1687,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/zlib/reports/'
            '20180907/linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    # The recent project entity that should be updated (|cov_info_project|).
    key = ndb.Key('CoverageInformation', 'zlib-20180907')
    expected_entity = data_types.CoverageInformation(
        date=datetime.date(2018, 9, 7),
        fuzzer='zlib',
        functions_covered=19,
        functions_total=47,
        edges_covered=987,
        edges_total=1687,
        html_report_url=(
            'https://storage.googleapis.com/oss-fuzz-coverage/zlib/reports/'
            '20180907/linux/index.html'))
    self.assertCoverageInformation(entities[key], expected_entity)
    del entities[key]

    # Should not have any entities left unverified. Ensures collect logic of
    # not creating duplicated entities if there is an existing one. In practice,
    # an existing entity could either be created by an earlier execution of
    # the cron task, or by the corpus pruning task.
    self.assertEqual(len(entities), 0)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Download files from GCS."""

import os
import urllib.parse

from flask import request

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import blobs
from handlers import base_handler
from libs import access
from libs import gcs
from libs import helpers
from libs.issue_management import issue_tracker_utils


class Handler(base_handler.Handler, gcs.SignedGcsHandler):
  """Download a file from GCS."""

  def _send_blob(self,
                 blob_info,
                 testcase_id,
                 is_minimized=False,
                 fuzzer_binary_name=None):
    """Send the blob."""
    minimized_string = 'minimized-' if is_minimized else ''
    fuzzer_binary_string = (
        fuzzer_binary_name + '-' if fuzzer_binary_name else '')
    if testcase_id:
      _, extension = os.path.splitext(blob_info.filename)
      filename = (
          'clusterfuzz-testcase-'
          '{minimized_string}{fuzzer_binary_string}{testcase_id}{extension}'
      ).format(
          minimized_string=minimized_string,
          fuzzer_binary_string=fuzzer_binary_string,
          testcase_id=testcase_id,
          extension=extension)
    else:
      filename = blob_info.filename

    content_disposition = str('attachment; filename=%s' % filename)
    response = self.serve_gcs_object(blob_info.bucket, blob_info.object_path,
                                     content_disposition)
    response.headers['Content-disposition'] = content_disposition
    return response

  def check_public_testcase(self, blob_info, testcase):
    """Check public testcase."""
    if blob_info.key() != testcase.minimized_keys:
      return False

    if not testcase.bug_information:
      return False

    issue_tracker = issue_tracker_utils.get_issue_tracker_for_testcase(testcase)
    issue = issue_tracker.get_issue(testcase.bug_information)
    if not issue:
      return False

    # If the issue is explicitly marked as view restricted to committers only
    # (OSS-Fuzz only), then don't allow public download.
    if 'restrict-view-commit' in issue.labels:
      return False

    return True

  def get(self, resource=None):
    """Handle a get request with resource."""
    testcase = None
    testcase_id = request.args.get('testcase_id')
    if not testcase_id and not resource:
      raise helpers.EarlyExitException('No file requested.', 400)

    if testcase_id:
      try:
        testcase = data_handler.get_testcase_by_id(testcase_id)
      except errors.InvalidTestcaseError:
        raise helpers.EarlyExitException('Invalid testcase.', 400)

      if not resource:
        if testcase.minimized_keys and testcase.minimized_keys != 'NA':
          resource = testcase.minimized_keys
        else:
          resource = testcase.fuzzed_keys

    fuzzer_binary_name = None
    if testcase:
      fuzzer_binary_name = testcase.get_metadata('fuzzer_binary_name')

    resource = str(urllib.parse.unquote(resource))
    blob_info = blobs.get_blob_info(resource)
    if not blob_info:
      raise helpers.EarlyExitException('File does not exist.', 400)

    if (testcase and testcase.fuzzed_keys != blob_info.key() and
        testcase.minimized_keys != blob_info.key()):
      raise helpers.EarlyExitException('Invalid testcase.', 400)

    if (utils.is_oss_fuzz() and testcase and
        self.check_public_testcase(blob_info, testcase)):
      # Public OSS-Fuzz testcase.
      return self._send_blob(
          blob_info,
          testcase.key.id(),
          is_minimized=True,
          fuzzer_binary_name=fuzzer_binary_name)

    is_minimized = testcase and blob_info.key() == testcase.minimized_keys
    if access.has_access():
      # User has general access.
      return self._send_blob(blob_info, testcase_id, is_minimized,
                             fuzzer_binary_name)

    # If this blobstore file is for a testcase, check if the user has access to
    # the testcase.
    if not testcase:
      raise helpers.AccessDeniedException()

    if access.can_user_access_testcase(testcase):
      return self._send_blob(blob_info, testcase_id, is_minimized,
                             fuzzer_binary_name)

    raise helpers.AccessDeniedException()
